"""KCH Driver."""
from __future__ import annotations

import atexit
import warnings
from enum import IntEnum, unique
from typing import Optional

try:
    import RPi.GPIO as GPIO  # isort: ignore
    HAS_HAT = True
except ImportError:
    HAS_HAT = False


@unique
class RobotLEDs(IntEnum):
    """Mapping of LEDs to GPIO Pins."""

    START = 9

    USER_A_RED = 24
    USER_A_GREEN = 10
    USER_A_BLUE = 25
    USER_B_RED = 27
    USER_B_GREEN = 23
    USER_B_BLUE = 22
    USER_C_RED = 4
    USER_C_GREEN = 18
    USER_C_BLUE = 17

    @classmethod
    def all_leds(cls) -> list[int]:
        """Get all LEDs."""
        return [c.value for c in cls]

    @classmethod
    def user_leds(cls) -> list[tuple[int, int, int]]:
        """Get the user programmable LEDs."""
        return [
            (cls.USER_A_RED, cls.USER_A_GREEN, cls.USER_A_BLUE),
            (cls.USER_B_RED, cls.USER_B_GREEN, cls.USER_B_BLUE),
            (cls.USER_C_RED, cls.USER_C_GREEN, cls.USER_C_BLUE),
        ]


@unique
class UserLED(IntEnum):
    """User Programmable LEDs."""

    A = 0
    B = 1
    C = 2


class Colour():
    """User LED colours."""

    OFF = (False, False, False)
    RED = (True, False, False)
    YELLOW = (True, True, False)
    GREEN = (False, True, False)
    CYAN = (False, True, True)
    BLUE = (False, False, True)
    MAGENTA = (True, False, True)
    WHITE = (True, True, True)


class KCH:
    """KCH Board."""
    __slots__ = ('_leds', '_pwm')

    def __init__(self) -> None:
        if HAS_HAT:
            GPIO.setmode(GPIO.BCM)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                # If this is not the first time the code is run this init will
                # cause a warning as the gpio are already initialized, we can
                # suppress this as we know the reason behind the warning
                GPIO.setup(RobotLEDs.all_leds(), GPIO.OUT, initial=GPIO.LOW)
            self._pwm: Optional[GPIO.PWM] = None
        self._leds = tuple(
            LED(rgb_led) for rgb_led in RobotLEDs.user_leds()
        )

        if HAS_HAT:
            # We are not running cleanup so the LED state persists after the code completes,
            # this will cause a warning with `GPIO.setup()` which we can suppress
            # atexit.register(GPIO.cleanup)

            # Cleanup just the start LED to turn it off when the code exits
            # Mypy isn't aware of the version of atexit.register(func, *args)
            atexit.register(GPIO.cleanup, RobotLEDs.START)  # type: ignore[call-arg]

    @property
    def _start(self) -> bool:
        """Get the state of the start LED."""
        return GPIO.input(RobotLEDs.START) if HAS_HAT else False

    @_start.setter
    def _start(self, value: bool) -> None:
        """Set the state of the start LED."""
        if HAS_HAT:
            if self._pwm:
                # stop any flashing the LED is doing
                self._pwm.stop()
                self._pwm = None
            GPIO.output(RobotLEDs.START, GPIO.HIGH if value else GPIO.LOW)

    def _flash_start(self) -> None:
        """Set the start LED flashing at 1Hz."""
        if HAS_HAT:
            self._pwm = GPIO.PWM(RobotLEDs.START, 1)
            self._pwm.start(50)

    @property
    def leds(self) -> tuple['LED', ...]:
        """User programmable LEDs."""
        return self._leds


class LED:
    """User programmable LED."""
    __slots__ = ('_led',)

    def __init__(self, led: tuple[int, int, int]):
        self._led = led

    @property
    def r(self) -> bool:
        """Get the state of the Red LED segment."""
        return GPIO.input(self._led[0]) if HAS_HAT else False

    @r.setter
    def r(self, value: bool) -> None:
        """Set the state of the Red LED segment."""
        if HAS_HAT:
            GPIO.output(self._led[0], GPIO.HIGH if value else GPIO.LOW)

    @property
    def g(self) -> bool:
        """Get the state of the Green LED segment."""
        return GPIO.input(self._led[1]) if HAS_HAT else False

    @g.setter
    def g(self, value: bool) -> None:
        """Set the state of the Green LED segment."""
        if HAS_HAT:
            GPIO.output(self._led[1], GPIO.HIGH if value else GPIO.LOW)

    @property
    def b(self) -> bool:
        """Get the state of the Blue LED segment."""
        return GPIO.input(self._led[2]) if HAS_HAT else False

    @b.setter
    def b(self, value: bool) -> None:
        """Set the state of the Blue LED segment."""
        if HAS_HAT:
            GPIO.output(self._led[2], GPIO.HIGH if value else GPIO.LOW)

    @property
    def colour(self) -> tuple[bool, bool, bool]:
        """Get the colour of the user LED."""
        if HAS_HAT:
            return (
                GPIO.input(self._led[0]),
                GPIO.input(self._led[1]),
                GPIO.input(self._led[2]),
            )
        else:
            return False, False, False

    @colour.setter
    def colour(self, value: tuple[bool, bool, bool]) -> None:
        """Set the colour of the user LED."""
        if not isinstance(value, (tuple, list)) or len(value) != 3:
            raise ValueError("The LED requires 3 values for it's colour")

        if HAS_HAT:
            GPIO.output(
                self._led,
                tuple(
                    GPIO.HIGH if v else GPIO.LOW for v in value
                ),
            )



from __future__ import annotations

import atexit
import json
import logging
from threading import Lock
from typing import Any, Callable, Optional, TypedDict, Union
from urllib.parse import urlparse

import paho.mqtt.client as mqtt

LOGGER = logging.getLogger(__name__)


class MQTTClient:
    def __init__(
        self,
        client_name: Optional[str] = None,
        topic_prefix: Optional[str] = None,
        mqtt_version: int = mqtt.MQTTv5,
        use_tls: Union[bool, str] = False,
        username: str = '',
        password: str = '',
    ) -> None:
        self._subscriptions_lock = Lock()
        self._subscriptions: dict[
            str, Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None]
        ] = {}
        self.topic_prefix = topic_prefix
        self._client_name = client_name

        self._client = mqtt.Client(client_id=client_name, protocol=mqtt_version)
        self._client.on_connect = self._on_connect

        if use_tls:
            self._client.tls_set()
            if use_tls == 'insecure':
                self._client.tls_insecure_set(True)

        if username:
            self._client.username_pw_set(username, password)

    def connect(self, host: str, port: int) -> None:
        """
        Connect to the MQTT broker and start event loop in background thread.

        Registers an atexit routine that tears down the client.
        """
        if self._client.is_connected():
            LOGGER.error("Attempting connection, but client is already connected.")
            return

        try:
            self._client.connect(host, port, keepalive=60)
        except (TimeoutError, ValueError, ConnectionRefusedError):
            LOGGER.error(f"Failed to connect to MQTT broker at {host}:{port}")
            return
        self._client.loop_start()
        atexit.unregister(self.disconnect)  # Avoid duplicate atexit handlers
        atexit.register(self.disconnect)

    @classmethod
    def establish(
        cls, host: str, port: int, **kwargs: Any,
    ) -> 'MQTTClient':
        """Create client and connect."""
        client = cls(**kwargs)
        client.connect(host, port)
        return client

    def disconnect(self) -> None:
        """Disconnect from the broker and close background event loop."""
        self._client.disconnect()
        self._client.loop_stop()
        atexit.unregister(self.disconnect)

    def subscribe(
        self,
        topic: str,
        callback: Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None],
        abs_path: bool = False,
    ) -> None:
        """Subscribe to a topic and assign a callback for messages."""
        if not abs_path and self.topic_prefix is not None:
            full_topic = f"{self.topic_prefix}/{topic}"
        else:
            full_topic = topic

        with self._subscriptions_lock:
            self._subscriptions[full_topic] = callback
        self._subscribe(full_topic, callback)

    def _subscribe(
        self,
        topic: str,
        callback: Callable[[mqtt.Client, Any, mqtt.MQTTMessage], None],
    ) -> None:
        LOGGER.debug(f"Subscribing to {topic}")
        self._client.message_callback_add(topic, callback)
        self._client.subscribe(topic, qos=1)

    def unsubscribe(self, topic: str) -> None:
        """Unsubscribe from a topic."""
        try:
            with self._subscriptions_lock:
                del self._subscriptions[topic]
        except KeyError:
            pass
        self._client.message_callback_remove(topic)
        self._client.unsubscribe(topic)

    def publish(
        self,
        topic: str,
        payload: Union[bytes, str],
        retain: bool = False,
        *,
        abs_topic: bool = False,
    ) -> None:
        """Publish a message to the broker."""
        if not self._client.is_connected():
            LOGGER.debug(
                "Attempted to publish message, but client is not connected.",
            )
            return

        if not abs_topic and self.topic_prefix:
            topic = f"{self.topic_prefix}/{topic}"

        try:
            self._client.publish(topic, payload=payload, retain=retain, qos=1)
        except ValueError as e:
            raise ValueError(f"Cannot publish to MQTT topic: {topic}") from e

    def wrapped_publish(
        self,
        topic: str,
        payload: Union[bytes, str],
        retain: bool = False,
        *,
        abs_topic: bool = True,
    ) -> None:
        """Wrap a payload up to be decodable as JSON."""
        if isinstance(payload, bytes):
            payload = payload.decode('utf-8')
        self.publish(
            topic,
            json.dumps({"data": payload}),
            retain=retain, abs_topic=abs_topic)

    def _on_connect(
        self, client: mqtt.Client, userdata: Any, flags: dict[str, int], rc: int,
        properties: Optional[mqtt.Properties] = None,
    ) -> None:
        """Callback run each time the client connects to the broker."""
        if rc != mqtt.CONNACK_ACCEPTED:
            LOGGER.warning(
                f"Failed to connect to MQTT broker. Return code: {mqtt.error_string(rc)}"
            )
            return

        LOGGER.debug("Connected to MQTT broker.")

        with self._subscriptions_lock:
            for topic, callback in self._subscriptions.items():
                self._subscribe(topic, callback)


class MQTTVariables(TypedDict):
    host: str
    port: int
    topic_prefix: str
    use_tls: Union[bool, str]
    username: Optional[str]
    password: Optional[str]


def unpack_mqtt_url(url: str) -> MQTTVariables:
    """
    Unpack an MQTT URL into its components to be passed to MQTTClient.

    The URL should be in the format:
    mqtt[s]://[<username>[:<password>]]@<host>[:<port>]/<topic_root>
    """
    url_parts = urlparse(url)

    if url_parts.scheme not in ('mqtt', 'mqtts'):
        raise ValueError(f"Invalid scheme: {url_parts.scheme}")

    use_tls = (url_parts.scheme == 'mqtts')

    if url_parts.hostname is None:
        raise ValueError("No hostname given")

    return MQTTVariables(
        host=url_parts.hostname,
        port=url_parts.port or (8883 if use_tls else 1883),
        topic_prefix=url_parts.path,
        use_tls=use_tls,
        username=url_parts.username,
        password=url_parts.password,
    )



"""General utility functions and classes for the sr-robot3 package."""
from __future__ import annotations

import logging
import signal
import socket
from abc import ABC, abstractmethod
from types import FrameType
from typing import Any, Mapping, NamedTuple, Optional, TypeVar

from serial.tools.list_ports import comports
from serial.tools.list_ports_common import ListPortInfo

T = TypeVar('T')
logger = logging.getLogger(__name__)


class BoardIdentity(NamedTuple):
    """
    A container for the identity of a board.

    All the board firmwares should return this information in response to
    the *IDN? query.

    :param manufacturer: The manufacturer of the board
    :param board_type: The short name of the board, i.e. PBv4B
    :param asset_tag: The asset tag of the board,
        this should match what is printed on the board
    :param sw_version: The firmware version of the board
    """
    manufacturer: str = ""
    board_type: str = ""
    asset_tag: str = ""
    sw_version: str = ""


class Board(ABC):
    """
    This is the base class for all boards.

    Slots are used to prevent adding attributes to the class at runtime.
    They must also be defined in subclasses.
    """
    __slots__ = ('_identity',)

    @staticmethod
    @abstractmethod
    def get_board_type() -> str:
        """
        The string that is expected to be returned by the board's firmware.

        This should match the value in the board_type field of the response to
        the *IDN? query. This is implemented as a static method because class-level
        attributes are not compatible with slots.

        :return: The board type name
        """
        pass  # pragma: no cover

    @abstractmethod
    def identify(self) -> BoardIdentity:
        """
        Return the board's identity as reported by the firmware.

        :return: The board's identity
        """
        pass  # pragma: no cover


def map_to_int(
        x: float,
        in_min: float,
        in_max: float,
        out_min: int,
        out_max: int,
) -> int:
    """
    Map a value from the input range to the output range, returning the value as an int.

    This is used to convert a float value to an integer value for sending to the board.

    :param x: The value to map
    :param in_min: The lower bound of the input range
    :param in_max: The upper bound of the input range
    :param out_min: The lower bound of the output range
    :param out_max: The upper bound of the output range
    :return: The mapped value
    """
    value = (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
    return int(value)


def map_to_float(
        x: int,
        in_min: int,
        in_max: int,
        out_min: float,
        out_max: float,
        precision: int = 3,
) -> float:
    """
    Map a value from the input range to the output range, returning the value as a float.

    This is used to convert an integer value from the board to a float value.

    :param x: The value to map
    :param in_min: The lower bound of the input range
    :param in_max: The upper bound of the input range
    :param out_min: The lower bound of the output range
    :param out_max: The upper bound of the output range
    :param precision: The number of decimal places to round the output to, defaults to 3
    :return: The mapped value
    """
    value = (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min
    return round(value, precision)


def singular(container: Mapping[str, T]) -> T:
    """
    Extract the single item from a container of one item.

    This is used to access individual boards without needing their serial number.

    :param container: A mapping of connected boards of a type
    :raises RuntimeError: If there is not exactly one of this type of board connected
    :return: _description_
    """
    length = len(container)

    if length == 1:
        return list(container.values())[0]
    elif length == 0:
        raise RuntimeError('No boards of this type found')
    else:
        raise RuntimeError(f'expected only one to be connected, but found {length}')


def obtain_lock(lock_port: int = 10653) -> socket.socket:
    """
    Bind to a port to claim it and prevent another process using it.

    We use this to prevent multiple robot instances running on the same machine,
    as they would conflict with each other.

    :param lock_port: The port to bind to, defaults to 10653 to be compatible with sr-robot3
    :raises OSError: If the port failed to bind, indicating another robot instance is running
    :return: The bound socket
    """
    lock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # We bind to a port to claim it and prevent another process using it
    try:
        lock.bind(("localhost", lock_port))
    except OSError:
        raise OSError('Unable to obtain lock, Is another robot instance already running?')

    return lock


def float_bounds_check(value: Any, min_val: float, max_val: float, err_msg: str) -> float:
    """
    Test that a value can be converted to a float and is within the given bounds.

    Errors are caught and re-raised with a more descriptive message.

    :param value: The value to test
    :param min_val: The minimum allowed value
    :param max_val: The maximum allowed value
    :param err_msg: The error message to raise if the value is invalid or out of bounds
    :raises TypeError: If the value cannot be converted to a float
    :raises ValueError: If the value is out of bounds
    :return: The input value converted to a float
    """
    try:
        new_val = float(value)
    except ValueError as e:
        raise TypeError(err_msg) from e

    if (new_val < min_val) or (new_val > max_val):
        raise ValueError(err_msg)

    return new_val


def get_USB_identity(port: ListPortInfo) -> BoardIdentity:
    """
    Generate an approximate identity for a board using the USB descriptor.

    This data will be overridden by the firmware once communication is established,
    but is used for early logging messages and error handling.

    :param port: The USB port information from pyserial
    :return: An initial identity for the board
    """
    try:
        return BoardIdentity(
            manufacturer=port.manufacturer or "",
            board_type=port.product or "",
            asset_tag=port.serial_number or "",
        )
    except Exception:
        logger.warning(
            f"Failed to pull identifying information from serial device {port.device}")
        return BoardIdentity()


def ensure_atexit_on_term() -> None:
    """
    Ensure `atexit` triggers on `SIGTERM`.
    > The functions registered via [`atexit`] are not called when the program is
      killed by a signal not handled by Python
    """

    if signal.getsignal(signal.SIGTERM) != signal.SIG_DFL:
        # If a signal handler is already present for SIGTERM,
        # this is sufficient for `atexit` to trigger, so do nothing.
        return

    def handle_signal(handled_signum: int, frame: Optional[FrameType]) -> None:
        """
        Handle the given signal by outputting some text and terminating the process.
        This will trigger `atexit`.
        """
        logger.info(signal.strsignal(handled_signum))
        exit(1)

    # Add the null-ish signal handler
    signal.signal(signal.SIGTERM, handle_signal)


def list_ports() -> None:
    """
    Print a list of all connected USB serial ports.

    Output is in the format:
    ```
    VID:PID - Manufacturer - Product - Serial Number
    ```
    """
    print("VID:PID - Manufacturer - Product - Serial Number")
    for serial_port in comports():
        if serial_port.vid and serial_port.pid:
            # Only print USB serial ports
            print(
                f"{serial_port.vid:04X}:{serial_port.pid:04X} - {serial_port.manufacturer} "
                f"- {serial_port.product} - {serial_port.serial_number}"
            )



"""Functions for killing the robot after a certain amount of time."""
import logging
import os
import signal
import sys
from threading import Timer
from types import FrameType
from typing import Optional

logger = logging.getLogger(__name__)


def timeout_handler(signal_type: int, stack_frame: Optional[FrameType]) -> None:
    """
    Handle the timeout to kill the current process.

    This function is called when the timeout expires and will stop the robot's main process.
    In order for this to work, any threads that are created must be daemons.

    NOTE: This function is not called on Windows.

    :param signal_type: The sginal that triggered this handler
    :param stack_frame: The stack frame at the time of the signal
    """
    logger.info("Timeout expired: Game Over!")
    exit(0)


def win_timeout_handler() -> None:
    """
    Kill the main process on Windows.

    This function is called when the timeout expires and will stop the robot's main process.
    In order for this to work, any threads that are created must be daemons.

    NOTE: This function is only called on Windows.
    """
    logger.info("Timeout expired: Game Over!")
    os.kill(os.getpid(), signal.SIGTERM)


def kill_after_delay(timeout_seconds: int) -> None:
    """
    Kill the robot after a certain amount of time.

    Interrupts main process after the given delay.

    :param timeout_seconds: The number of seconds to wait before killing the robot
    """

    if sys.platform == "win32":
        # Windows doesn't have SIGALRM,
        # so we approximate its functionality using a delayed SIGTERM
        timer = Timer(timeout_seconds, win_timeout_handler)
        timer.start()
    else:
        signal.signal(signal.SIGALRM, timeout_handler)
        signal.alarm(timeout_seconds)
    logger.info(f"Timeout set: {timeout_seconds}s")



"""Test the timeout function."""
from time import sleep, time

import os
import signal
import subprocess
import sys
import pytest
from unittest.mock import Mock

from sr.robot3.timeout import kill_after_delay


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on Windows")
def test_kill_after_delay() -> None:
    """Test that the process is killed within the time."""
    with pytest.raises(SystemExit):
        kill_after_delay(2)
        sleep(3)


@pytest.mark.skipif(sys.platform != "win32", reason="only runs on Windows")
def test_kill_after_delay_windows(monkeypatch) -> None:
    """Test that the process is killed within the time on windows."""
    kill = Mock()
    pid = os.getpid()
    monkeypatch.setattr(os, "kill", kill)

    kill_after_delay(2)
    sleep(1)

    kill.assert_not_called()

    sleep(1.5)
    kill.assert_called_once_with(pid, signal.SIGTERM)


def test_kill_after_delay_e2e() -> None:
    start_time = time()
    child = subprocess.Popen([
        sys.executable,
        "-c",
        'from sr.robot3.timeout import kill_after_delay; '
        'import time; kill_after_delay(2); time.sleep(10)',
    ], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    child.wait(timeout=5)

    assert time() - start_time == pytest.approx(2, rel=1)

    if sys.platform == "win32":
        # Windows terminates uncleanly
        assert child.returncode == signal.SIGTERM
    else:
        assert child.returncode == 0
